/*-----------------09-07-94 09:36am-----------------
 * This is include file "declares" a generic singly linked list. It is to be
 * used in conjunction with the file 'slist.c', which "defines" a generic
 * singly linked list. To use it, define "TYPE" for the element you wish the
 * SList to contain, and then include this file. In order for an element to
 * be containable, it must be a struct. You must ensure that it has some
 * necessary fields, and certain functions must be declared and perform
 * important operations.
 * All the necessary code to create "declarations" for a SList of 'element' is
 * shown below. Note that all fields, and functions shown are essential for
 * the SList to work, except 'elementInit'. The 'elementInit' function is
 * advisable but not necessary.


	typedef struct element {
		struct TYPE *next;  // next element in the list
		...other fields of "element"
	} element;

	void elementInit(element *e,...other parameters);
	//initializes element "e"

	void elementDone(element *e);
	//destroys element "e"

	void elementCopy(element *e,element *o);
	//copys element "o" to element "e"

	#define TYPE element
	#include "slist.h"


 * This effectively automaticaly creates the following function "declarations".


	typedef struct elementSList {
		element *next;
	} elementSList;

	void elementSListInit(elementSList *l);
	//Initializes a element list

	void elementSListDone(elementSList *l);
	//destroys a element list, freeing any included elements

	void elementSListCopy(elementSList *l,elementSList *o);
	//copys element list 'o' into element list 'l'

	void elementSListRem(elementSList *l,element *e);
	//removes element 'e' from the list 'l'

	void elementSListAdd(elementSList *l,element *e);
	//adds element 'e' to the start of list 'l'

	void elementSListApp(elementSList *l,element *e);
	//appends element 'e' to the end of list 'l'

	element *elementSListPop(elementSList *l);
	//returns and pops a element from the top of the list 'l'

	element *elementSListPul(elementSList *l);
	//returns and puls a element from the bottom of the list 'l'

	void elementSListAft(elementSList *l,element *p,element *e);
	//inserts element 'e' after 'p' in list 'l'.
	//If 'p'=NULL, 'e'is put at the front of the list

	void elementSListBef(elementSList *l,element *p,element *e);
	//inserts element 'e' before 'p' in list 'l'.
	//If 'p'=NULL, 'e'is put at the end of the list

	TYPE *elementSListPrev(elementSList *l,element *e);
	//returns the element before 'e' in the list 'l'.
	//If 'e'=NULL, returns the end of the list

 * --------------------------------------------------*/

/*-----------------09-07-94 09:36am-----------------
 * Added features: Turning off unused functions.
 *
 * Every function declaration, apart from SListInit and SListDone can now
 * be turned off, to 'hide' or remove any unused functions. This is done
 * by defining a name "SLISTxxx_OFF" for each feature disabled. Note that
 * these constants are are undef'ed at the end of this file, so if you
 * have multiple SList instances in the one file, you must ensure all the
 * required disabling names are defined before this file is included.
 * Note that SListPrev cannot be turned off as it is needed by other routines.
 * Here is an example:

	#define SLISTCOPY_OFF
	#define SLISTAPP_OFF
	#define TYPE  element
	#include "slist.h"

	#define SLISTCOPY_OFF
	#define SLISTAPP_OFF
	#define TYPE  element
	#include "slist.c"

 * This declares and defines an elementSList as above, but without any
 * elementSListCopy or elementSListApp functions. Note that the disabling
 * #defines are required before both including slist.h AND including
 * slist.c
 * --------------------------------------------------*/

/*-----------------01/02/96 17:02-------------------
 * Added features: the FUNC(n) definition is no longer required.
 *
 * I now know how to trick the pre-processor to do the job by using
 * "double indirection" with JOIN and JOIN2 (see below).
 * --------------------------------------------------*/


#ifndef TYPE
#error TYPE needs to be defined
#endif

#define FUNC(f)   JOIN(TYPE,f)
#define JOIN(x,y) JOIN2(x,y)
#define JOIN2(x,y) x##y

#define TYPESList FUNC(SList)


typedef struct TYPESList {
	TYPE *next;
} TYPESList;


void FUNC(SListInit)(TYPESList *l);
//Initializes a TYPE list

void FUNC(SListDone)(TYPESList *l);
//destroys a TYPE list, freeing any included TYPEs


#ifndef SLISTCOPY_OFF
void FUNC(SListCopy)(TYPESList *l,TYPESList *o);
//copys TYPE list 'o' into TYPE list 'l'
#endif //SLISTCOPY_OFF


#ifndef SLISTREM_OFF
void FUNC(SListRem)(TYPESList *l,TYPE *e);
//removes TYPE 'e' from the list 'l'
#endif //SLISTREM_OFF


#ifndef SLISTADD_OFF
void FUNC(SListAdd)(TYPESList *l,TYPE *e);
//adds TYPE 'e' to the start of list 'l'
#endif //SLISTADD_OFF


#ifndef SLISTAPP_OFF
void FUNC(SListApp)(TYPESList *l,TYPE *e);
//appends TYPE 'e' to the end of list 'l'
#endif //SLISTAPP_OFF


#ifndef SLISTPOP_OFF
TYPE *FUNC(SListPop)(TYPESList *l);
//returns and pops a TYPE from the top of the list 'l'
#endif //SLISTPOP_OFF


#ifndef SLISTPUL_OFF
TYPE *FUNC(SListPul)(TYPESList *l);
//returns and pulls a TYPE from the bottom of the list 'l'
#endif //SLISTPUL_OFF


#ifndef SLISTAFT_OFF
void FUNC(SListAft)(TYPESList *l,TYPE *p,TYPE *e);
//inserts TYPE 'e' after 'p' in list 'l'.
//If 'p'=NULL, 'e'is put at the front of the list
#endif //SLISTAFT_OFF


#ifndef SLISTBEF_OFF
void FUNC(SListBef)(TYPESList *l,TYPE *p,TYPE *e);
//inserts TYPE 'e' before 'p' in list 'l'.
//If 'p'=NULL, 'e'is put at the end of the list
#endif //SLISTBEF_OFF


TYPE *FUNC(SListPrev)(TYPESList *l,TYPE *e);
//returns the TYPE before 'e' in the list 'l'.
//If 'e'=NULL, returns the end of the list


#undef SLISTCOPY_OFF
#undef SLISTREM_OFF
#undef SLISTADD_OFF
#undef SLISTAPP_OFF
#undef SLISTPOP_OFF
#undef SLISTPUL_OFF
#undef SLISTAFT_OFF
#undef SLISTBEF_OFF

#undef FUNC
#undef TYPE
#undef TYPESList
