/* squirm.c

   Version: squirm-1.0betaB

   See the README file for usage instructions, and INSTALL
   for help on installing
   
        Last change:  DTB   8 Jul 98    5:11 pm
*/

   
/*
  Squirm - A Redirector for Squid

  Maintained by Chris Foote, chris@senet.com.au
  Copyright (C) 1998 Chris Foote & Wayne Piekarski
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
    
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  Please see the file GPL in this directory for full copyright
  information.
*/


#include"squirm.h"
#include"paths.h"
#include"log.h"

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<unistd.h>
#include<sys/signal.h>
#include<ctype.h>


/* Declare pattern_itemSList type and methods*/
#define TYPE pattern_item
#define SLISTCOPY_OFF
#include "slist.h"

/* Define pattern_itemSList type and methods*/
#define TYPE pattern_item
#define SLISTCOPY_OFF
#include "slist.c"


/* This is the main pattern list */
pattern_itemSList pattern_list;


/* Patterns initializer */
void PatternsInit(void)
{
        pattern_itemSListInit(&pattern_list);
}


/* Patterns destroyer */
void PatternsDone(void)
{
        pattern_itemSListDone(&pattern_list);
}


/* Patterns loader */
int PatternsLoad(char *config_file)
{
        pattern_item *e = NULL;
        char buff[MAX_BUFF];
        FILE *fp;

        fp = fopen(config_file, "rt");
        if(fp == NULL) {
                log(LOG_ERROR, "unable to open config file\n");
                return 0;
        }

        while( fgets(buff, MAX_BUFF, fp) ) {

                /* trim newline */
                if ( strlen(buff) > 1 ) {
                        buff[strlen(buff) - 1] = '\0';
                }

                /* skip blank lines and comments */
                if ( buff[0] == '#' || buff[0] == '\0' )
                        continue;

                /* allocate pattern_item */
                e = (pattern_item *)malloc(sizeof(pattern_item));
                if ( e == NULL ) {
                        log(LOG_ERROR, "unable to allocate memory in PatternsLoad()\n");
                        fclose(fp);
                        return 0;
                }

                /* initialise pattern_item and add*/
                if ( pattern_itemInit(e,buff) ) {
                        pattern_itemSListApp(&pattern_list, e);
                } else{
                        log(LOG_ERROR, "unable to initialise pattern in PatternsLoad()\n");
                        free(e);
                        fclose(fp);
                        return 0;
                }
        }

        fclose(fp);
        return 1;
}


/* Patterns match to url
   returns replacement URL for a match, NULL if no match found */
char *PatternsMatch(char *url)
{
        pattern_item *e = NULL;
        char *replacement = NULL;

        for(e = pattern_list.next; e != NULL; e = e->next) {
                replacement = pattern_itemMatch(e,url);
                if ( replacement != NULL ) {
                        return (replacement);
                }
        }
        return (NULL);
}




/* load the stdin for the redirector into
   an IN_BUFF structure 
   Sets in_buff.url to "" if the fields can't
   be converted
*/
int load_in_buff(char *buff, IN_BUFF *in_buff)
{
        int converted;

        strcpy(in_buff->url, "");
        strcpy(in_buff->src_address, "");
        strcpy(in_buff->ident, "");
        strcpy(in_buff->method, "");

        converted = sscanf(buff, "%s %s %s %s\n", in_buff->url, 
                           in_buff->src_address, in_buff->ident,
                           in_buff->method);
        if(converted != 4) {
                log(LOG_ERROR, "%d = incorrect number of scanned fields\n", converted);
                return 1;
        }

#ifdef DEBUG
        log(LOG_DEBUG, "Request:%s %s %s %s\n", in_buff->url, in_buff->src_address,
            in_buff->ident, in_buff->method);
#endif

        /* all methods must be GET */
        if((strcmp(in_buff->method, "") == 0) || (strcmp(in_buff->method, "GET"))) {
#ifdef DEBUG
                log(LOG_DEBUG, "in_buff.method not \"GET\" : %s\n", in_buff->src_address);
#endif
                return 1;
        }

        return 0;
}




/* pattern_item initializer */
int pattern_itemInit(pattern_item *e, char *pattern)
{
        char type[MAX_BUFF];
        char first[MAX_BUFF];
        char second[MAX_BUFF];
        char accel[MAX_BUFF];
        int parenthesis;
        int replacement;
        int stored;
        int regex_flags;

        assert(e!=NULL);
#ifndef NDEBUG
        e->next=NULL;
#endif//NDEBUG
#ifdef DEBUG
        log(LOG_DEBUG, "Adding pattern: [%s]\n", pattern);
#endif//DEBUG
        
        e->pattern = NULL;
        e->replacement = NULL;
        e->accel = NULL;
        e->type = NORMAL;
        e->accel_type = ACCEL_OFF;

        stored = sscanf(pattern, "%s %s %s %s", type, first, second, accel);
        if((stored < 3) || (stored > 4)) {
                log(LOG_ERROR, "invalid pattern for line: [%s]\n", pattern);
                return 0;
        }

        parenthesis = count_parenthesis (first);
        regex_flags = REG_EXTENDED | REG_NOSUB;
        if ( !strcmp(second, "-") ) {
                strcpy(second, "");
        }
        if ( !strcmp(type, "regex") ) {
                replacement = count_replacement (second);
                e->case_sensitive = 1;
        } else if ( !strcmp(type, "regexi") ) {
                replacement = count_replacement (second);
                e->case_sensitive = 0;
                regex_flags |= REG_ICASE;
        } else if ( !strcmp(type, "redir") ) {
                replacement = 0;
                e->case_sensitive = 1;
                e->type = REDIRECT;
        } else if ( !strcmp(type, "rediri") ) {
                replacement = 0;
                e->case_sensitive = 0;
                regex_flags |= REG_ICASE;
                e->type = REDIRECT;
        } else {
                log (LOG_ERROR, "invalid type in pattern: [%s]\n", type);
                return 0;
        }

        if ( replacement > 0 ) {
                e->type = EXTENDED;
                regex_flags ^= REG_NOSUB;
        }

        if ( regcomp(&e->cpattern, first, regex_flags) ) {
                log(LOG_ERROR, "Invalid regex in pattern: [%s]\n", first);
                return 0;
        }

        if ( replacement < 0 ) {
                log (LOG_ERROR, "invalid replacement in pattern: [%s]\n", second);
                regfree(&e->cpattern);
                return 0;
        }

        e->pattern = strdup(first);
        if(e->pattern == NULL) {
                log(LOG_ERROR, "unable to allocate memory in pattern_itemInit()\n");
                regfree(&e->cpattern);
                return 0;
        }

        e->replacement = strdup(second);
        if(e->replacement == NULL) {
                log(LOG_ERROR, "unable to allocate memory in pattern_itemInit()\n");
                regfree(&e->cpattern);
                free(e->pattern);
                return 0;
        }

        /* use accelerator string if it exists */
        if(stored == 4) {
                e->accel = get_accel(accel, &e->accel_type, e->case_sensitive);
                if(e->accel == NULL) {
                        log(LOG_ERROR, "unable to allocate memory in pattern_itemInit()\n");
                        regfree(&e->cpattern);
                        free(e->pattern);
                        free(e->replacement);
                        return 0;
                }
        }
        return 1;
}


/* pattern_item destroyer */
void pattern_itemDone(pattern_item *e)
{
        assert(e!=NULL);

        free(e->pattern);
        free(e->replacement);
        if(e->accel)
            free(e->accel);
        regfree(&e->cpattern);
}


/* pattern_item match to url
  returns replacement url for a match, NULL if no match found */
char *pattern_itemMatch(pattern_item *e, char *url)
{
        char *replacement = NULL;

        /* check for accelerator string and return NULL if no match*/
        if ( e->accel_type ) {
                if ( !match_accel(url, e->accel, e->accel_type, e->case_sensitive) )
                    return (NULL);
#ifdef DEBUG
                log(LOG_DEBUG, "URL [%s] matched accelerator [%s]\n", url, e->accel);
#endif
        }

        /* check pattern type */
        if ( e->type == NORMAL ) {
                /* NORMAL, return replacement string if matched */
                if ( regexec(&e->cpattern, url, 0, 0, 0) == 0 ) {
                        replacement = strdup(e->replacement);
#ifdef DEBUG
                        log(LOG_DEBUG, "URL [%s] matched NORMAL pattern [%s]\n", url, e->pattern);
#endif
                }
        } else if (e->type == EXTENDED) {
                /* EXTENDED, perform regex substitition */
                replacement = regexsub(&(e->cpattern), url, e->replacement);
#ifdef DEBUG
                log(LOG_DEBUG, "URL [%s] matched EXTENDED pattern [%s]\n", url, e->pattern);
#endif
        }

        return (replacement);
}




/* parses an accelerator string */
/* returns the stripped accelerator string or NULL 
 if memory can't be allocated 
 converts the accel string to lower case
 if(case_sensitive) */
char *get_accel(char *accel, int *accel_type, int case_sensitive) 
{

        int len, i;
        char *new_accel = NULL;

        len = strlen(accel);
        if(accel[0] == '^') {
                *accel_type = ACCEL_START;
        } else if(accel[len - 1] == '$') {
                *accel_type = ACCEL_END;
        } else {
                *accel_type = ACCEL_NORMAL;
        }
        
        if(*accel_type == ACCEL_START || *accel_type == ACCEL_END) {

                /* copy the strings */
                new_accel = (char *)malloc(sizeof(char) * strlen(accel));
                if(new_accel == NULL)
                    return NULL;

                if(*accel_type == ACCEL_START) {
                        if(case_sensitive)
                            for(i = 0; i < len; i++)
                                new_accel[i] = accel[i+1];
                        else
                            for(i = 0; i < len; i++)
                                new_accel[i] = tolower(accel[i+1]);
                }

                if(*accel_type == ACCEL_END) {
                        if(case_sensitive)
                            for(i = 0; i < len - 1; i++)
                                new_accel[i] = accel[i];
                        else
                            for(i = 0; i < len - 1; i++)
                                new_accel[i] = tolower(accel[i]);
                        new_accel[i] = '\0';
                }

        } else {

                new_accel = strdup(accel);
                if(!case_sensitive) {
                        for(i = 0; i < len; i++)
                            new_accel[i] = tolower(accel[i]);
                        new_accel[i] = '\0';
                }
        }
        return new_accel;
}



/* matches accelerator to url */
/* return 1 if url contains accel */
int match_accel(char *url, char *accel, int accel_type, int case_sensitive)
{
        int i, offset;
        char l_accel[MAX_BUFF];
        int accel_len;
        int url_len;

        if(accel_type == ACCEL_NORMAL) {
                if(case_sensitive) {
                        for(i = 0; url[i] != '\0'; i++)
                            l_accel[i] = url[i];
                } else {
                        /* convert to lower case */
                        for(i = 0; url[i] != '\0'; i++)
                            l_accel[i] = tolower(url[i]);
                }
                l_accel[i] = '\0';

                if(strstr(url, l_accel)) {
                        return 1;
                } else {
                        return 0;
                }
        }

        if(accel_type == ACCEL_START) {
                accel_len = strlen(accel);
                url_len = strlen(url);
                if(url_len < accel_len)
                    return 0;

                if(case_sensitive) {
                        for(i = 0; i < accel_len; i++) {
                                if(accel[i] != url[i])
                                    return 0;
                        }
                } else {
                        for(i = 0; i < accel_len; i++) {
                                if(accel[i] != tolower(url[i]))
                                    return 0;
                        }
                }
                return 1;
        }

        if(accel_type == ACCEL_END) {
                accel_len = strlen(accel);
                url_len = strlen(url);
                offset = url_len - accel_len;

                if(offset < 0)
                    return 0;

                if(case_sensitive) {
                        for(i = 0; i < accel_len; i++) {
                                if(accel[i] != url[i+offset])
                                    return 0;
                        }
                } else {
                        for(i = 0; i < accel_len; i++) {
                                if(accel[i] != tolower(url[i+offset]))
                                    return 0;
                        }
                }
                return 1;
        }

        /* we shouldn't reach this section! */
        return 0;
}




/* regular expression match and replace
  returns NULL on no match or failed substitution. */
char *regexsub(regex_t *preg, char *str, char *rep)
{
        regmatch_t match_data[10];
        char buffer[MAX_BUFF];
        char *in_ptr;
        char *out_ptr;
        int replay_num;
        int count;
        char *replacement;

        /* Perform the regex call */
        if (regexec (preg, str, 10, match_data, 0) != 0)
            return (NULL);

        /* setup the traversal pointers */
        in_ptr = rep;
        out_ptr = buffer;

        /* Traverse the replacement string */
        while ( *in_ptr != '\0' ) {

                if ( *in_ptr == '\\' ) {
                        /* We have a '\', inc the in pointer to the next char*/
                        in_ptr++;

                        if ( isdigit( *in_ptr ) ) {
                                /* We have a digit following, replace it */

                                /* Convert the current in_ptr into a number */
                                replay_num = (*in_ptr) - '0';

                                /* Now copy the chars from the replay string */
                                for (count = match_data[replay_num].rm_so;
                                     count < match_data[replay_num].rm_eo;
                                     count++) {
                                        *out_ptr = str[count];
                                        out_ptr++;
                                }

                        } else if ( *in_ptr == '\0' ) {
                                /* We have a null following a '\' */
                                return (NULL);

                        } else {
                                /* Otherwise, insert as a literal */
                                *out_ptr = *in_ptr;
                                out_ptr++;
                        }

                } else {
                        /* Its not a backslash, just put it in */
                        *out_ptr = *in_ptr;
                        out_ptr++;
                }

                /* Increment the in pointer to the next character*/
                in_ptr++;
        }

        /* Terminate the string */
        *out_ptr = '\0';

        /* The return string is done, return to the caller */
        replacement = strdup(buffer);
        return (replacement);
}


/* counts parenthesis in regex pattern */
int count_parenthesis (char *pattern)
{
        int lcount = 0;
        int rcount = 0;
        char *reg_ptr;

        /* setup the traversal pointers */
        reg_ptr = pattern;

        /* Traverse the regex string looking for '(' and ')' */
        while ( *reg_ptr != '\0' ) {

                if ( *reg_ptr == '\\' ) {
                        /* We have found a '\', lets skip it and ignore what follows */
                        reg_ptr++;
                        if ( *reg_ptr == '\0')
                            /* what follows is a null, we have '\' at the end of the regex */
                            return -1;
                } else if ( *reg_ptr == '(' ) {
                        /* We have a '(', lets count it*/
                        lcount++;
                } else if ( *reg_ptr == ')' ) {
                        /* We have a ')', lets count it*/
                        rcount++;
                }

                /* Increment the pointer to the next character */
                reg_ptr++;
        }

        /* Check that left == right */
        if (lcount != rcount)
            return (-1);

        return (lcount);
}
        

/* counts substitutions in regex replacement string */
int count_replacement (char *replace)
{
        int count = 0;
        char *reg_ptr;

        /* setup the traversal pointer */
        reg_ptr = replace;

        /* Traverse the replacement string looking for "\n" */
        while ( *reg_ptr != '\0' ) {

                if ( *reg_ptr == '\\' ) {
                        /* We have found a '\', lets skip it and check what follows */
                        reg_ptr++;
                        if ( isdigit( *reg_ptr ) ) {
                                /* We have a digit following, lets count it*/
                                count++;
                        } else if ( *reg_ptr == '\0') {
                                /* what follows is a null, we have '\' at the end of the replacement */
                                return -1;
                        }
                }

                /* Increment the pointer to the next character */
                reg_ptr++;
        }

        return (count);
}

