/* squirm.h */

/*
  Squirm - A Redirector for Squid
  
  Maintained by Chris Foote, chris@senet.com.au
  Copyright (C) 1998 Chris Foote & Wayne Piekarski
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  Please see the file GPL in this directory for full copyright
  information.
	Last change:  DTB   8 Jul 98    5:13 pm
*/


#ifndef SQUIRM_H
#define SQUIRM_H

#include<stdarg.h>
#include<sys/types.h>
#include<regex.h>


void PatternsInit(void);
void PatternsDone(void);
int  PatternsLoad(char *config_file);
char *PatternsMatch(char *url);


#define MAX_BUFF 8000

typedef struct IN_BUFF {
	char url[MAX_BUFF];
	char src_address[MAX_BUFF];
	char ident[MAX_BUFF];
	char method[MAX_BUFF];
} IN_BUFF;

int load_in_buff(char *buff, IN_BUFF *in_buff);


#define NORMAL   1
#define EXTENDED 2
#define REDIRECT 3

#define ACCEL_OFF    0
#define ACCEL_NORMAL 1
#define ACCEL_START  2
#define ACCEL_END    3

typedef struct pattern_item {
	struct pattern_item *next;
	char *pattern;
	char *replacement;
	char *accel;
	int type;
	int case_sensitive;
	int accel_type;
	regex_t cpattern;
} pattern_item;

int  pattern_itemInit(pattern_item *e, char *pattern);
void pattern_itemDone(pattern_item *e);
//void pattern_itemCopy(pattern_item *e,pattern_item *o);
char *pattern_itemMatch(pattern_item *e, char *url);

char *get_accel(char *accel, int *accel_type, int case_sensitive);
int match_accel(char *url, char *accel, int accel_type, int case_sensitive);


char *regexsub(regex_t *preg, char *str, char *rep);
int count_parenthesis (char *pattern);
int count_replacement (char *replace);

#endif

